#include <exec/tasks.h>
#include <exec/ports.h>
#include <exec/memory.h>
#include <exec/lists.h>
#include <exec/semaphores.h>
#include <exec/execbase.h>
#include <exec/alerts.h>
#include <exec/libraries.h>
#include <exec/interrupts.h>
#include <exec/resident.h>
#include <graphics/gfxbase.h>
#include <intuition/classes.h>
#include <intuition/classusr.h>
#include <intuition/cghooks.h>
#include <intuition/gadgetclass.h>
#include <intuition/intuitionbase.h>
#include <dos/dosextens.h>
#include <datatypes/datatypes.h>
#include <datatypes/datatypesclass.h>
#include <datatypes/pictureclass.h>

#include <datatypes/pictureclassext.h>
#include <cybergraphics/cybergraphics.h>


#include <emul/emulinterface.h>
#include <emul/emulregs.h>

#include <ppcinline/exec.h>
#include <ppcinline/datatypes.h>
#include <ppcinline/graphics.h>
#include <ppcinline/intuition.h>
#include <ppcinline/dos.h>
#include <ppcinline/utility.h>

#include <public/quark/quark.h>
#include <public/proto/quark/syscall_protos.h>


struct LibBase
{
	struct Library		 Lib;
	BPTR			 SegList;
	struct	ExecBase	*SysBase;
	struct GfxBase		*GfxBase;
	struct IntuitionBase	*IntuitionBase;
	struct DosLibrary	*DOSBase;
	struct Library		*SuperClassBase;
	struct Library		*DataTypesBase;
	struct Library		*UtilityBase;
	struct Library		*IFFParseBase;

	Class			*Class;
	struct SignalSemaphore	Semaphore;
};

#define	SysBase		MyLibBase->SysBase
#define	GfxBase		MyLibBase->GfxBase
#define	IntuitionBase	MyLibBase->IntuitionBase
#define	DOSBase		MyLibBase->DOSBase
#define	SuperClassBase	MyLibBase->SuperClassBase
#define	DataTypesBase	MyLibBase->DataTypesBase
#define	UtilityBase	MyLibBase->UtilityBase
#define	IFFParseBase	MyLibBase->IFFParseBase


#define CLASSNAME	(MyLibBase->Lib.lib_Node.ln_Name)


struct FFRHandle
 {
  struct StandardPacket fh_Pkt;
  struct MsgPort *fh_ReplyPort;
  struct FileHandle *fh_Handle;
  UBYTE *fh_Buffer,*fh_Pre,*fh_Ptr;
  LONG fh_BufferSize,fh_Left;
  BOOL fh_Flag;
 };

LONG	ReadPBM(Class*,
                Object*,
                struct TagItem*);

void	CloseFFR(struct LibBase*,
                 struct FFRHandle *);
struct FFRHandle  *OpenFFR(struct LibBase*,
                           BPTR);

LONG	FFRGetC(struct LibBase*,
                struct FFRHandle *);
LONG	FFRRead(struct LibBase*,
                struct FFRHandle *,
                APTR,
                ULONG);

#define READ(h,p,b) (FFRRead(MyLibBase,h,p,b)!=(b))
#define MySetIoErr(e) ((struct Process *)FindTask(NULL))->pr_Result2=(e)

Class	*LIB_CreateClass(struct LibBase		*MyLibBase);
ULONG	LIB_DeleteClass(struct LibBase		*MyLibBase,
                        Class			*MyClass);



ULONG	CallHookA(struct Hook	*MyHook,
                  Object	*MyObject,
		  void		*MyMsg);

ULONG	DoMethodA(Object	*MyObject,
                  Msg		MyMsg);

ULONG	DoSuperMethodA(Class		*MyClass,
                       Object		*MyObject,
                       Msg		MyMsg);

ULONG	CoerceMethodA(Class		*MyClass,
                      Object		*MyObject,
                      Msg		MyMsg);


#define CallHook(MyHook, MyObject, tags...) \
	({ULONG _tags[] = { tags }; CallHookA((MyHook), (MyObject), (APTR)_tags);})

#define DoMethod(MyObject, tags...) \
	({ULONG _tags[] = { tags }; DoMethodA((MyObject), (APTR)_tags);})

#define DoSuperMethod(MyClass, MyObject, tags...) \
	({ULONG _tags[] = { tags }; DoSuperMethodA((MyClass), (MyObject), (APTR)_tags);})

#define CoerceMethod(MyClass, MyObject, tags...) \
	({ULONG _tags[] = { tags }; CoerceMethodA((MyClass), (MyObject), (APTR)_tags);})
